﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;

/* --- © 2022 JaxZoa <JaxZoa@gmail.com> --- */

namespace JaxPlugins
{

    public class StepCycle
    {
        public CLS manager;

        protected List<Step> steps;

        protected float stepSpacingIdle;
        protected float stepSpacingMoving;
        protected float stepSpacingKneelingIdle;
        protected float stepSpacingKneelingMoving;

        protected float groundOffset;
        protected float groundOffsetKneeling;

        protected float stepSpacingRandom_Min;
        protected float stepSpacingRandom_Max;

        protected float stepDurationRandom_Min;
        protected float stepDurationRandom_Max;

        protected bool stepInPlace;
        protected float stepInPlaceAmplitude;
        protected float stepInPlaceAmplitudeOffset;

        protected bool isIdle;

        public StateChange idleStateDelegate;
        public delegate void StateChange(bool _newState);

        protected int activeStep = 0;

        public StepCycle(CLS _manager)
        {
            manager = _manager;
        }

        public void SetSteps(List<Step> _steps)
        {
            steps = _steps;
        }

        public void Init()
        {
            foreach(Step _step in steps)
            {
                _step.Init(ComputeStepTargetPos(_step), manager.TargetTransform.rotation);
            }
        }

        bool _idleState;

        public void ComputeSteps()
        {
            if(!steps[activeStep].IsActive)
            {
                _idleState = false;
                
                steps[activeStep].StartStep(ComputeStepTargetPos(steps[activeStep]), manager.TargetTransform.rotation);

                foreach (Step _step in steps)
                {
                    if (_step.IsIdle)
                        _idleState = true;
                }

                if(_idleState != isIdle)
                {
                    isIdle = _idleState;

                    if(isIdle)
                        idleStateDelegate.Invoke(isIdle);
                }
            }
            else
            {
                if (steps[activeStep].PlayedThisCycle)
                    activeStep++;

                if (activeStep >= steps.Count)
                {
                    foreach(Step _step in steps)
                    {
                        _step.ResetStep();
                    }

                    activeStep = 0;
                }
            }
        }

        RaycastHit[] hitInfos = new RaycastHit[100];
        int hitCount = 0;
        float stepSpacing = 0;
        Vector3 stepTargetPos;
        Vector3 stepRayCastPos;
        float _stepHeight;

        protected Vector3 ComputeStepTargetPos(Step _step)
        {
            stepTargetPos = manager.TargetTransform.position;
            stepTargetPos.y = _step.TargetPos.y - (manager.IsStanding ? groundOffset : groundOffsetKneeling);

            _stepHeight = -10000;

            if (manager.IsStanding)
            {
                stepSpacing = Mathf.Lerp(stepSpacingIdle, stepSpacingMoving, manager.SpeedFactor);               
            }
            else
            {
                stepSpacing = Mathf.Lerp(stepSpacingKneelingIdle, stepSpacingKneelingMoving, manager.SpeedFactor);                
            }

            if (stepInPlace)
                stepSpacing += UnityEngine.Random.Range(stepSpacingRandom_Min, stepSpacingRandom_Max) * StepInPlaceAmplitude;
            else if (!manager.IsIdle)
                stepSpacing += UnityEngine.Random.Range(stepSpacingRandom_Min, stepSpacingRandom_Max);

            if (!_step.Right)
                stepSpacing = -stepSpacing;

            stepTargetPos += manager.TargetTransform.right * stepSpacing;

            stepRayCastPos = stepTargetPos;

            hitCount = Physics.SphereCastNonAlloc(stepRayCastPos + Vector3.up * 0.8f, 0.05f, -Vector3.up, hitInfos, 10f);

            if (hitCount > 0)
            {
                for(int i = 0; i < hitCount; i++)
                {
                    if (hitInfos[i].collider == null)
                        continue;

                    if (!manager.GroundColliders.Contains(hitInfos[i].collider))
                        continue;


                    if (hitInfos[i].point.y > _stepHeight)
                    {
                        _stepHeight = hitInfos[i].point.y;
                    }
                }
            }

            if (_stepHeight == -10000)
                _stepHeight = stepTargetPos.y;

            _stepHeight += manager.IsStanding ? groundOffset : groundOffsetKneeling;

            stepTargetPos.y = _stepHeight;

            return stepTargetPos;
        }

        public Step GetStep(int _stepIndex)
        {
            if (_stepIndex < 0 || _stepIndex >= steps.Count)
                return null;

            return steps[_stepIndex];
        }

        public void SetAllStepsDurations(float _newDuration)
        {
            foreach(Step _s in steps)
            {
                _s.Duration = _newDuration;
            }
        }

        public void ResetStepTiming()
        {
            foreach (Step _s in steps)
            {
                _s.ResetStep();
                activeStep = 0;
            }
        }

        public bool IsIdle
        {
            get { return isIdle; }
        }

        public float StepSpacingIdle
        {
            get { return stepSpacingIdle; }

            set { stepSpacingIdle = value; }
        }

        public float StepSpacingMoving
        {
            get { return stepSpacingMoving; }

            set { stepSpacingMoving = value; }
        }

        public float StepSpacingKneelingIdle
        {
            get { return stepSpacingKneelingIdle; }

            set { stepSpacingKneelingIdle = value; }
        }

        public float StepSpacingKneelingMoving
        {
            get { return stepSpacingKneelingMoving; }

            set { stepSpacingKneelingMoving = value; }
        }

        public float StepSpacingRandom_Min
        {
            set { stepSpacingRandom_Min = value; }
            get { return stepSpacingRandom_Min; }
        }

        public float StepSpacingRandom_Max
        {
            set { stepSpacingRandom_Max = value; }
            get { return stepSpacingRandom_Max; }
        }

        public float StepDurationRandom_Min
        {
            set { stepDurationRandom_Min = value; }
            get { return stepDurationRandom_Min; }
        }

        public float StepDurationRandom_Max
        {
            set { stepDurationRandom_Max = value; }
            get { return stepDurationRandom_Max; }
        }

        public float GroundOffset
        {
            get { return groundOffset; }

            set { groundOffset = value; }
        }

        public float GroundOffsetKneeling
        {
            get { return groundOffsetKneeling; }

            set { groundOffsetKneeling = value; }
        }

        public float StepInPlaceAmplitude
        {
            get { return stepInPlaceAmplitude + stepInPlaceAmplitudeOffset; }

            set { stepInPlaceAmplitude = value; }
        }

        public float StepInPlaceAmplitudeOffset
        {
            get { return stepInPlaceAmplitudeOffset; }

            set { stepInPlaceAmplitudeOffset = value; }
        }

        public bool StepInPlace
        {
            get { return stepInPlace; }

            set { stepInPlace = value; }
        }


    }
}